/****************************************************************************
 *                                                                          *
 * Filename:    dnsrh.c                                                     *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                      *
 * Created:     15/10/03                                                    *
 * Modified:    26/01/04                                                    *
 * Description: A DNS Results Hash, stores maps between IPv6 addrs and v4   *
 *              reserved addresses.                                         *
 *                                                                          *
 * This module author was inspired by the following open code and           *
 * interesting writings:                                                    *
 *    http://www.xml.com/ldd/chapter/book/                                  *
 *    http://www.dirac.org/linux/writing/lkmpg/lkmpg.html                   *
 *    http://en.tldp.org/LDP/khg/HyperNews/get/devices/devices.html         *
 *    http://jungla.dit.upm.es/~jmseyas/linux/kernel/hackers-docs.html      *
 *    http://www.kernelhacking.org/docs/kernelhacking-HOWTO/index.html      *
 *                                                                          *
 * (c) 2003 Tabitha Tipper, licensed under the GNU General Public License      *
 *     See the LICENSE file for details.                                    *
 ****************************************************************************/

/* FIXME: if the DNSRH isn't loaded then it returns 0.0.0.0 as an address...
 *        meaning all your traffic becomes broadcast
 */

/* FIXME: breaks under load, use get/put from userspace functions */

/* #define DEBUG_YES */

/* needed ones */
/* #include <linux/config.h> */
#include <linux/module.h>
#include <linux/kernel.h>

/* Don't use versioning, it breaks a lot of functions:
 * #include <linux/version.h> */
#include <linux/init.h>

/* #include <linux/slab.h> */

/* if versioning being used */
/*
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif  
*/

/* The character device definitions are here */
#include <linux/fs.h>

/* doesn't do much now, will later according to LKML */
#include <linux/wrapper.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
#include <asm/uaccess.h>  /* for get_user and put_user */
#endif

/* Include our own headers */
#include "dnsrh.h"
#include "dnsrh-support.c"

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

MODULE_AUTHOR("Tabitha Tipper <t.tipper@lancaster.ac.uk>");
MODULE_DESCRIPTION("A hash that stores IPv6 addresses as class E IPv4 for consistent reference");
MODULE_LICENSE("GPL");

/* The name for our device, as it will appear in 
 * /proc/devices */
#define DEVICE_NAME "dnsrh"

static ssize_t device_read(struct file *filp, 
		char *buff, size_t count, loff_t *offp)
{
	return 0;
}


static ssize_t device_write(struct file *filp, const char *buff,
	size_t count, loff_t *offp)
{
	return 0;
}

static int device_open(struct inode * inode, struct file * file)
{
	return 0;
}

/* this kludge deals with 2.2 versions of the kernel */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
static int device_release(struct inode * inode, struct file * file)
#else
static void device_release(struct inode * inode, struct file * file)
#endif
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
	return 0;
#endif
}

/***
 * The call that actually handles all the IOCTL stuff defined in the header
 * There is no need to check permissions here as those are done at the
 * filesystem level (don't do silly checks where they aren't needed).
 ***/
int device_ioctl(
	struct inode * inode,
	struct file * file,
	unsigned int ioctl_num,		/* The number of the ioctl          */
	unsigned long ioctl_param)	/* The pointer to the optional data */
{

	#ifdef DEBUG_YES
	printk(KERN_INFO "dnsrh: DEBUG: ioctl_num: %u, ioctl_param: %u\n", ioctl_num, (int) ioctl_param);
	#endif

	int err = 0;

	switch(ioctl_num) {

		/* will add (or not if already there) a v6 addr to the hash */
		/* returns 0 if it was there, 1 if it wasn't and 2 if it has
		 * to overwrite a v4 */
		case IOCTL_ADD_V6: 
			err = dnsrh_addV6Addr((ip_pair *) ioctl_param);
			return err;
			break;

		/* returns a v4 addr looked up from the v6 in the pair
		 * Returns 0 on success and fills the v4, returns 1 on fail */
		case IOCTL_CHECK_FOR_V4:
			err = dnsrh_checkForV4((ip_pair *) ioctl_param);
			return err;
			break;

		/* returns a v6 addr looked up from the v4 in the pair
		 * returns 0s and fills the .v6 on success, 1 on failure */
		case IOCTL_CHECK_FOR_V6:
			err = dnsrh_checkForV6((ip_pair *) ioctl_param);
			return err;
			break;

		default:
			return 1;

	}

	return 1;
}

/***
 * This struct contains the functions that will be called for each type of
 * operation on this device.  There are some you *CANNOT* afford not to have
 * in a Unix system.
 ***/
struct file_operations Fops = {
 	llseek:  NULL,
 	read:  device_read,
	write:  device_write,
	ioctl:  device_ioctl,
	open:  device_open,
	release: device_release,
};

/***
 * This is the function that kickstarts the module, setting up the hash
 ***/
static int __init dnsrh_init(void) 
{
	int err;

	/* Register the character device (atleast try) */
	/*err = module_register_chrdev(MAJOR_NUM, DEVICE_NAME, &Fops);*/
	err = register_chrdev(MAJOR_NUM, DEVICE_NAME, &Fops);

	/* Negative values signify an error */
	if (err < 0) {
		printk(KERN_INFO "dnsrh: Registering the character device failed with %d\n", err);
		return err;
	}

	/* Remember if you haven't create a device with:
	 * "mknod DEVICE_FILE_NAME c MAJOR_NUM 0" 
	 */


	/* Perform any setup thats needed by the hash table */
	err = dnsrh_initialise_hash();

	if (err != 0) { 
		printk(KERN_INFO "dnsrh: Failure to initialise hash: %d", err);
		return err;
	}

	printk(KERN_INFO "dnsrh: DNS Results Hash loaded.\n"); 
	return 0;
}

/***
 * This function stops the hash and carefully cleans the hash away again
 ***/
static void __exit dnsrh_exit(void) 
{ 
	int err;

	/* Unregister the device */
	/* err = module_unregister_chrdev(MAJOR_NUM, DEVICE_NAME); */
	err = unregister_chrdev(MAJOR_NUM, DEVICE_NAME);
 
	/* If there's an error, report it */ 
	if (err < 0) { 
		printk(KERN_INFO "dnsrh: Error in module_unregister_chrdev: %d\n", err);
	}

	dnsrh_clean_hash();

	printk(KERN_INFO "dnsrh: DNS Results Hash unloaded.\n");
}

module_init(dnsrh_init)
module_exit(dnsrh_exit)
