/****************************************************************************
 *                                                                          *
 * Filename: dnsrh.h                                                        *
 * Author:   Tabitha Tipper <t.tipper@lancaster.ac.uk>                         *
 * Created:  26/10/03                                                       *
 * Modified: 25/01/04                                                       *
 * Description: IOCTL call definitions, available here for both client libs *
 *              and the module itself, also functions from dnsrh-support.c  *
 *              Oh and structs too, ip_pair and single_v4                   *
 *                                                                          *
 * (c) 2003 Tabitha Tipper, licensed under the GNU General Public License      *
 *     See the LICENSE file for details.                                    *
 ****************************************************************************/

/* FIXME: Stop all functions being exported, only export what we need to
 * prevent polluting Linus's glorious kernel
 */

#ifndef dnsrh_h
#define dnsrh_h

#include <linux/ioctl.h>
#include <linux/slab.h>
#include <asm/semaphore.h>

#define MAXSIZE 500	/* The maximum size for the hash */

/***
 * The ip_pair struct, contains an IPv4 address and an IPv6 address.
 * Also has an optional pointer to the next address in the chain, if there is
 * no next address this variable should be set to NULL.
 ***/
typedef struct _ip_pair { 
	unsigned char v4[4];
	unsigned char v6[16];
	struct _ip_pair * next;
} ip_pair;

/***
 * A struct that contains a single V4 address, for when _ip_pair is overkill
 ***/
typedef struct _single_v4 { 
	unsigned char a;
	unsigned char b;
	unsigned char c;
	unsigned char d;
} single_v4;

/***
 * Major device number, must be known to all involved so the correct drivers
 * can be loaded.
 *
 * 61 is in the middle of the 60-63 range of experiement/local major devices
 ***/
#define MAJOR_NUM 61

/***
 * IOCTL_ADD_V6 is a call that takes an IPv4 addy as the filled in part of 
 * an ip_pair and adds it to the hash (and obviously if its already there 
 * it does nothing, but this saves time on checking).
 *
 * Returns a 0 if freshly added, 1 if already there and 2 if a v4 was
 * overwritten.
 ***/
#define IOCTL_ADD_V6 _IOR(MAJOR_NUM, 0, ip_pair *)

/***
 * IOCTL_CHECK_FOR_V4 accepts a V6 address and returns the V4
 * corresponding to it.  Of failing that fills the struct with 0s.
 *
 * returns a 0 on success, 1 on failure.
 ***/
#define IOCTL_CHECK_FOR_V4 _IOWR(MAJOR_NUM, 1, ip_pair *)

/***
 * IOCTL_CHECK_FOR_V6 accepts a V4 address and returns the V6
 * corresponding to it.  Of failing that fills the struct with 0s.
 *
 * returns a 0 on success, 1 on failure.
 ***/
#define IOCTL_CHECK_FOR_V6 _IOWR(MAJOR_NUM, 2, ip_pair *)

/* The name of the device file */
#define DEVICE_FILE_NAME "dnsrh"


/***
 * initialise_hash performs any initialisations that need to be performed.
 * Ideally this should mean that other people can write drop in replacements
 * for the atrocious mess that I've written here and the rest of the code
 * (thats much better, honest) won't notice at all.
 *
 * It returns 0 on succes, 1 on failure
 ***/
int dnsrh_initialise_hash();

/***
 * Cleans the hash properly
 *
 * FIXME: Often causes NULL dereferencing Kernel panics.... rewrite.
 *
 ***/
int dnsrh_clean_hash();


/***
 * Compares two V6 parts of two ip_pair structs
 * 
 * returns 0 if they're identical and 1 if not
 ***/
int dnsrh_compareV6(ip_pair a, ip_pair b);

/***
 * Compares two V4 parts of two ip_pair structs
 * 
 * returns 0 if they're identical and 1 if not
 ***/
int dnsrh_compareV4(ip_pair a, ip_pair b);


/***
 * Accepts a char pointer to a V4 address.
 * returns a hash key in the range (0 - MAXSIZE).
 ***/
int dnsrh_hashV4Addr(ip_pair in);

/***
 * Accepts a char pointer to a V6 address.
 * returns a hash key in the range (0 - MAXSIZE).
 ***/
int dnsrh_hashV6Addr(ip_pair in);

/*** 
 * Selects the next V4 from the programs list available and returns it.
 *
 * In the event that you run out of V4s it should loop around to the beginning
 * again.
 *
 * According to RFC 3330 the "reserved for future use" block is 240.0.0.0/4,
 * This function will use the range 240.0.0.0-240.255.255.254
 *
 * If the returned address is 0.0.0.0 then an error has occured.
 ***/
single_v4 dnsrh_generateV4();

/***
 * This function accepts an ip_pair struct containing an IPv6 address.  It
 * adds it to the Hash (if its already added it does nothing).
 *
 * It modifies the struct it was given to contain the matching V4 address.
 *
 * This function returns 0 if the address was already in the hash, 1 if its
 * been freshly added to the hash, 2 if a v4 address was overwritten and
 * -ERESTARTSYS in event of a mutex error.
 ***/
int dnsrh_addV6Addr(ip_pair * in);

/***
 * Accepts an ip_pair struct with an IPv6 address in, looks up the
 * corresponding V4 address and fills the struct.
 *
 * If there is no corresponding V4 address it fills the in.v4[x] parts with
 * 0s and returns 1.
 *
 * If there was a V4 address is returns 0 and fills the struct correctly.
 ***/
int dnsrh_checkForV4(ip_pair * in);

/***
 * Accepts an ip_pair struct wtih an IPv4 address in it, it looks up the
 * corresponding V6 address and fills the struct.
 *
 * If there is no corresponding V6 address it fills the in.v6[x] parts with
 * 0s and returns 1;
 *
 * If there was a V6 address it returns 0 and fills the struct correctly.
 ***/
int dnsrh_checkForV6(ip_pair * in);


#endif
