/***************************************************************************
 *                                                                         *
 * Filename:    v4tv6pm.c                                                  *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                     *
 * Created:     07/01/04                                                   *
 * Modified:    02/02/04                                                   *
 * Description: A netfilter kernel module, rewrites outgoing packets to    *
 *              from IPv4 to IPv6 before putting them back on the wire.    *
 *                                                                         *
 * (c) 2004 Tabitha Tipper, licensed under the GNU General Public License     *
 *     See the LICENSE file for details.                                   *
 ***************************************************************************/

/* BADCLASS
 *
 * include/linux/net/in.h      <-- tells us that Class E is bad.
 *
 * See docs/what-i-did for details.
 */

/*
 * see /usr/src/linux/include/linux/skbuff.h for struct skbuff info
 * see /usr/src/linux/include/linux/ip.h for iphdr
 * see /usr/src/linux/include/linux/ipv6.h for ipv6hdr
 *
 * see /usr/src/linux/include/linux/in6.h for struct in6_addr
 */

#define DEBUG_YES

/* needed ones */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>


/* For netfilter */
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv6.h>

/* For playing with skbuffs */
#include <linux/skbuff.h>
#include <net/ipv6.h>
#include <linux/ip.h>
#include <linux/ipv6.h>
#include <net/checksum.h>


/* maybe */
#include <linux/if_ether.h>

/* definitely but who knows whats in each */
#include <linux/udp.h>
#include <linux/icmp.h>
#include <net/icmp.h>
#include <net/ip.h>
#include <net/tcp.h>
#include <net/route.h>
#include <net/ip6_route.h>
#include <linux/netfilter_ipv4/ip_tables.h>


/* for versioning */
/*
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif  
*/
/* doesn't do much now, will later */
#include <linux/wrapper.h>

/* my own stuff */
#include "iopm.h"
#include "../dnsrh/dnsrh.h"

#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) ((a)*65536+(b)*256+(c))
#endif

MODULE_AUTHOR("Tabitha Tipper <t.tipper@lancaster.ac.uk>");
MODULE_DESCRIPTION("A netfilter module that rewrites outgoing v4 packets to v6");
MODULE_LICENSE("GPL");


/***
 * A dummy function with functionality ripped from "ip6_maybe_reroute" in
 * ip6_output.c 
 *
 * Does what that does... kinda, intended to appease the gods of NF_HOOK as
 * this will get called for the return value of nf_hook_slow()
 ***/
int v4tv6pm_dummy(struct sk_buff * skb)
{
	return skb->dst->output(skb);
}

/***
 * hex-to-int
 *
 * Accepts a char and converts it from its charness to the value of that
 * symbol in hex notation as a decimal.
 *
 * Return -1 on error, 0-15 in other cases.
 ***/
int v4tv6pm_hti(char in) 
{ 
	switch (in) { 
		case '0':
			return 0;
		case '1':
			return 1;
		case '2':
			return 2;
		case '3':
			return 3;
		case '4':
			return 4;
		case '5':
			return 5;
		case '6':
			return 6;
		case '7':
			return 7;
		case '8':
			return 8;
		case '9':
			return 9;
		case 'a':
		case 'A':
			return 10;
		case 'b':
		case 'B':
			return 11;
		case 'c':
		case 'C':
			return 12;
		case 'd':
		case 'D':
			return 13;
		case 'e':
		case 'E':
			return 14;
		case 'f':
		case 'F':
			return 15;
		default:
			return -1;
	}

	return -1;
}


/* copies the char[] representation of v6 (normally from LOCAL_V6 to the
 * in6_addr struct
 */
void v4tv6pm_fixaddr(char * src, ip_pair dst, struct ipv6hdr * out) 
{ 
	int i = 0;
	int j = 0;

	for (i = 0; i < 16; i++) { 
		/* composite of two bits of char array */
		out->saddr.in6_u.u6_addr8[i] = 
			(__u8) ( (out->saddr.in6_u.u6_addr8[i] 
				  | v4tv6pm_hti(src[j]) ) << 4) & 0xF0;

		out->saddr.in6_u.u6_addr8[i] += 
			(__u8) (out->saddr.in6_u.u6_addr8[i]
				| v4tv6pm_hti(src[j + 1])) & 0x0F;

		j += 2;
		out->daddr.in6_u.u6_addr8[i] = (__u8) dst.v6[i];
	}
}

/***
 * This function deals with incoming packets and is named in the v4tv6_ops
 * struct
 ***/
static unsigned int
v4tv6_hook(unsigned int hook,
		struct sk_buff **pskb,
		const struct net_device *indev,
		const struct net_device *outdev,
		int (*okfn)(struct sk_buff *))
{

	/* FIXME: Q: where do we get src addr from?  A: from iopm.h */

	/* could use (*pskb)->pkt_type == PACKET_OUTGOING but that
	 * seems to match nothing.
	 */
	if ( (*pskb)->protocol == __constant_htons(ETH_P_IP)) { /* IPv4 */

		ip_pair * addr = NULL;
		int err = 0;
		struct iphdr * temp = (*pskb)->nh.iph;

		/* #ifdef DEBUG_YES
		   printk(KERN_INFO "v4tv6pm: 0. Gotten hold of packet.\n");
#endif */

		/* setup hash */
		addr = kmalloc(sizeof(ip_pair), GFP_KERNEL);

		for (err = 0; err < 4; err++) 
			addr->v4[err] = 0;

		for (err = 0; err < 16; err++) 
			addr->v6[err] = 0;

		addr->next = NULL;

		/* This shamelessly stolen from kernel.h
		 * Thanks for the code.
		 */
		addr->v4[0] = ((unsigned char *) &temp->daddr)[0];
		addr->v4[1] = ((unsigned char *) &temp->daddr)[1];
		addr->v4[2] = ((unsigned char *) &temp->daddr)[2];
		addr->v4[3] = ((unsigned char *) &temp->daddr)[3];

		/* #ifdef DEBUG_YES
		   printk(KERN_INFO "v4tv6pm: 0. about to lookup.\n");
#endif */

		/* perform lookup */
		err = dnsrh_checkForV6(addr);

		/* #ifdef DEBUG_YES
		   printk(KERN_INFO "v4tv6pm: 0. looked up: %d\n", err);
#endif */

		/* There was a v6, rewrite */
		if (0 == err) { 
			char * local = { LOCAL_V6 };
			struct ipv6hdr * v6 = NULL;
			struct sk_buff * newskb = NULL;
			struct flowi * flows = NULL;
			struct packet_type * pkt = NULL;
			struct net_device * newdev = NULL;

			/* #ifdef DEBUG_YES
			   printk(KERN_INFO "v4tv6pm: 1. mangling packet.\n");
#endif */

			/* Ensure we're the only ones playing here...
			 * shouldn't this be GFP_KERNEL not _ATOMIC? */

			/* All SKB ops which alloc must be _ATOMIC */
			*pskb = skb_unshare(*pskb, GFP_ATOMIC); 

			/* Some debugging code */
/*			if ((*pskb)->sk == NULL) {
				printk(KERN_INFO "v4tv6pm: 1. (*pskb)->sk null.\n");
			}
			else { 
				printk(KERN_INFO "v4tv6pm: 1. (*pskb)->sk not null.\n");
			}

			if ((*pskb)->dst == NULL) { 
				printk(KERN_INFO "v4tv6pm: 1. (*pskb)->dst null.\n");
			}
			else { 
				printk(KERN_INFO "v4tv6pm: 1. (*pskb)->dst not null.\n");
			}


			printk(KERN_INFO "v4tv6pm: (*pskb)-> data: %p, head: %p,len: %d", (*pskb)->data, (*pskb)->head, (*pskb)->len);

*/

			/* END OF Some debugging code */


			v6 = (struct ipv6hdr *) kmalloc(sizeof(struct ipv6hdr), GFP_KERNEL);

			/* newskb = skb_copy_expand((*pskb), *((*pskb)->head), *((*pskb)->tail), GFP_ATOMIC); */
			newskb = skb_copy_expand(*pskb, (((*pskb)->dst->dev->hard_header_len + 15) & ~15), skb_tailroom(*pskb), GFP_ATOMIC);

			newskb = skb_unshare(newskb, GFP_ATOMIC);

			flows = (struct flowi *) kmalloc(sizeof(struct flowi), GFP_KERNEL);

			pkt = (struct packet_type *) kmalloc(sizeof(struct packet_type), GFP_KERNEL);
			newdev = (struct net_device *) kmalloc(sizeof(struct net_device), GFP_KERNEL);



			/* JUST CHECKING */
/*			if (newskb->sk == NULL) {
						printk(KERN_INFO "v4tv6pm: 1. newskb->sk null.\n");
						}
						else { 
						printk(KERN_INFO "v4tv6pm: 1. newskb->sk not null.\n");
						}
*/

/*			if (newskb->dst == NULL) { 
						printk(KERN_INFO "v4tv6pm: 1. newskb->dst null.\n");
						}
						else { 
						printk(KERN_INFO "v4tv6pm: 1. newskb->dst not null.\n");
						}

						if (newskb->dst->neighbour == NULL) { 
						printk(KERN_INFO "v4tv6pm: 1. newskb->dst.neighbour is null.\n");
						}
						else { 
						printk(KERN_INFO "v4tv6pm: 1. newskb->dst.neighbour is not null.\n");
						}

						printk(KERN_INFO "v4tv6pm: newskb data: %p, head: %p, len: %d", newskb->data, newskb->head, newskb->len);
*/
			/* END OF JUST CHECKING */




			/* Change skbuff protocol type */
			newskb->protocol = 
				__constant_htons(ETH_P_IPV6);

			/* create new ipv6hdr */
			v6->priority = 0;
			v6->version = 6;

			v6->flow_lbl[0] = 0;
			v6->flow_lbl[1] = 0;
			v6->flow_lbl[2] = 0;

			v6->payload_len = (temp->tot_len - temp->ihl);

			/* Check for fragmentation, the v4->id and
			 * v4->frag_offset need to be turned into this */
			v6->nexthdr = 0;

			v6->hop_limit = temp->ttl;

			/* do ip_summed too */

			/* Transfer addresses over */
			v4tv6pm_fixaddr(local, *addr, v6);

			/* #ifdef DEBUG_YES
			   printk(KERN_INFO "v4tv6pm: 1. adjusting .nh\n");
#endif */
			/* setup new pointer */
			kfree(newskb->nh.iph);
			/* newskb->nh.iph = NULL; */

			newskb->nh.ipv6h = v6;


			/* Adjust the length for the new header, er.. we don't
			 * seem to do this... */
			/* newskb->len = (newskb->len + 20); */

			/* Fix the checksum */
			newskb->csum = 0;
			newskb->csum = csum_partial((*pskb)->head, (*pskb)->truesize, 0);


			/* ifdef DEBUG_YES
			   printk(KERN_INFO "v4tv6pm: 1. now constructing flows.\n");
#endif */

			/* Construct the 'struct flowi' */
			/*			
						if (newskb->sk != NULL) {
						flows->proto = newskb->sk->protocol;
						printk(KERN_INFO "v4tv6pm: ->proto\n");
						}
						else { 
						flows->proto = 0;
						printk(KERN_INFO "v4tv6pm: ->proto set to 0\n");
						}



						flows->nl_u.ip6_u.daddr = &newskb->nh.ipv6h->daddr;
						printk(KERN_INFO "v4tv6pm: ->daddr\n");
						flows->nl_u.ip6_u.saddr = &newskb->nh.ipv6h->saddr;
						printk(KERN_INFO "v4tv6pm: ->saddr\n");
						flows->nl_u.ip6_u.flowlabel = (__u32) newskb->nh.ipv6h->flow_lbl;
						printk(KERN_INFO "v4tv6pm: ->flowlabel\n");


						if (newskb->sk != NULL) { 
						flows->oif = newskb->sk->bound_dev_if;
						printk(KERN_INFO "v4tv6pm: ->oif\n");


						if (newskb->sk->protocol == IPPROTO_TCP) { 
						printk(KERN_INFO "v4tv6pm: ->tcp\n");
						flows->uli_u.ports.sport = newskb->h.th->source;
						flows->uli_u.ports.dport = newskb->h.th->dest;
						} 

						else if (newskb->sk->protocol == IPPROTO_UDP) {
						printk(KERN_INFO "v4tv6pm: ->udp\n");
						flows->uli_u.ports.sport = newskb->h.uh->source;
						flows->uli_u.ports.dport = newskb->h.uh->dest;
						}

						else if (newskb->sk->protocol == IPPROTO_ICMP) { 
						printk(KERN_INFO "v4tv6pm: ->icmp\n");
						flows->uli_u.icmpt.type = newskb->h.icmph->type;
						flows->uli_u.icmpt.code = newskb->h.icmph->code;
						}
						} 
						else { 
						printk(KERN_INFO "v4tv6pm: uli_u not set.\n");
						}

*/
			/* end of flowi construction */

			/* #ifdef DEBUG_YES
			   printk(KERN_INFO "v4tv6pm: 1. fixing checksum.\n");
#endif */


			/* EARLY RETURN! WARNING! DANGER DANGER! */

#ifdef DEBUG_YES
			printk(KERN_INFO "v4tv6pm: 1. about to do ip6_output\n");
#endif
			err = ip6_output(newskb);

			/* PREVIOUS ATTEMPTS HERE */
			/* err = ip6_xmit(newskb->sk, newskb, flows, 0); */
			/* err = NF_HOOK(PF_INET6, NF_IP6_LOCAL_OUT, newskb, NULL, outdev, ip6_route_output); */
			
			
			
			/* err = NF_HOOK(PF_INET6, 
					NF_IP6_POST_ROUTING, 
					newskb, 
					indev, 
					outdev, 
					v4tv6pm_dummy); 
*/

#ifdef DEBUG_YES
			printk(KERN_INFO "v4tv6pm: 1. ip6_output returned: %d\n", err);
#endif

			/* According to Henrik Nordstrom <hno@marasystems.com>
			 * We should be using: ipv6_rcv()
			 */

			/* remove us from our list */
			/* newskb->list = 0; */

			/* construct a packet_type struct */
/*			pkt->type = __constant_htons(ETH_P_IPV6);
			pkt->dev = NULL;
			pkt->func = NULL;
			pkt->data = NULL;
			pkt->next = NULL;
*/

			/* copy the netdevice */
/*			for (err = 0; err < IFNAMSIZ; err++) 
				newdev->name[err] = indev->name[err];
			
			newdev->rmem_end = indev->rmem_end;
			newdev->rmem_start = indev->rmem_start;
			newdev->mem_end = indev->mem_end;
			newdev->base_addr = indev->base_addr;
			newdev->irq = indev->irq;

			newdev->if_port = indev->if_port;
			newdev->dma = indev->dma;
			newdev->state = indev->state;
			
			newdev->next = indev->next;
			newdev->init = indev->init;
			newdev->next_sched = indev->next_sched;

			newdev->ifindex = indev->ifindex;
			newdev->iflink = indev->iflink;

			newdev->get_stats = indev->get_stats;
			newdev->get_wireless_stats = indev->get_wireless_stats;

			newdev->wireless_handlers = indev->wireless_handlers;
			newdev->ethtool_ops = indev->ethtool_ops;

			newdev->trans_start = indev->trans_start;
			newdev->last_rx = indev->last_rx;
			newdev->flags = indev->flags;
			newdev->gflags = indev->gflags;
			newdev->priv_flags = indev->priv_flags;

			newdev->unused_alignment_fixer = indev->unused_alignment_fixer;

			newdev->mtu = indev->mtu;
			newdev->type = indev->type;
			newdev->hard_header_len = indev->hard_header_len;
			newdev->priv = indev->priv;
			newdev->master = indev->master;
		
			for (err = 0; err < MAX_ADDR_LEN; err++) 	
				newdev->broadcast[err] = indev->broadcast[err];

			for (err = 0; err < MAX_ADDR_LEN; err++) 
				newdev->dev_addr[err] = indev->dev_addr[err];


			newdev->addr_len = indev->addr_len;

			newdev->mc_list = indev->mc_list;
			newdev->mc_count = indev->mc_count;

			newdev->promiscuity = indev->promiscuity;
			newdev->allmulti = indev->allmulti;

			newdev->watchdog_timeo = indev->watchdog_timeo;
			newdev->watchdog_timer = indev->watchdog_timer;

			newdev->atalk_ptr = indev->atalk_ptr;
			newdev->ip_ptr = indev->ip_ptr;
			newdev->dn_ptr = indev->dn_ptr;
			newdev->ip6_ptr = indev->ip6_ptr;
			newdev->ec_ptr = indev->ec_ptr;

			newdev->poll_list = indev->poll_list;
			newdev->quota = indev->quota;
			newdev->weight = indev->weight;

			newdev->qdisc = indev->qdisc;
			newdev->qdisc_sleeping = indev->qdisc_sleeping;
			newdev->qdisc_list = indev->qdisc_list;
			newdev->qdisc_ingress = indev->qdisc_ingress;
			newdev->tx_queue_len = indev->tx_queue_len;

			newdev->xmit_lock = indev->xmit_lock;
			newdev->xmit_lock_owner = indev->xmit_lock_owner;
			newdev->queue_lock = indev->queue_lock;
			newdev->refcnt = indev->refcnt;

			newdev->deadbeaf = indev->deadbeaf;
			newdev->features = indev->features;

			newdev->uninit = indev->uninit;
			newdev->destructor = indev->destructor;
			newdev->open = indev->open;
			newdev->stop = indev->stop;
			newdev->hard_start_xmit = indev->hard_start_xmit;
			newdev->poll = indev->poll;
			newdev->hard_header = indev->hard_header;
			newdev->rebuild_header = indev->rebuild_header;
			newdev->set_multicast_list = indev->set_multicast_list;
			newdev->set_mac_address = indev->set_mac_address;
			newdev->do_ioctl = indev->do_ioctl;
			newdev->set_config = indev->set_config;
			newdev->hard_header_cache = indev->hard_header_cache;
			newdev->header_cache_update = indev->header_cache_update;
			newdev->change_mtu = indev->change_mtu;
			newdev->tx_timeout = indev->tx_timeout;
			newdev->vlan_rx_register = indev->vlan_rx_register;
			newdev->vlan_rx_add_vid = indev->vlan_rx_add_vid;
			newdev->vlan_rx_kill_vid = indev->vlan_rx_kill_vid;
			newdev->hard_header_parse = indev->hard_header_parse;
			newdev->neigh_setup = indev->neigh_setup;
			newdev->accept_fastpath = indev->accept_fastpath;

			newdev->owner = indev->owner;
			newdev->br_port = indev->br_port;
			
#ifdef CONFIG_NET_FASTROUTE
			newdev->fastpath_lock = indev->fastpath_lock;
			newdev->fastpath = indev->fastpath;
#endif

#ifdef CONFIG_NET_DIVERT
			newdev->divert = indev->divert;
#endif */
			

			/* and finally send it on its way */
/*			ipv6_rcv(newskb, newdev, pkt); */

			kfree(addr);
			kfree(flows);
			kfree(pkt);
			kfree(newdev);
			/* kfree_skb(newskb); */
#ifdef DEBUG_YES
			printk(KERN_INFO "Packet 1. mangled, returning NF_DROP.\n");
#endif
			return NF_DROP;
		} 
		else { 
			/* #ifdef DEBUG_YES
			   printk(KERN_INFO "v4tv6pm: 2. not mangling packet.  err: %d\n", err);
#endif */
		}
		/* let modified packet continue */
		kfree(addr);
	}
	/* #ifdef DEBUG_YES
	   printk("v4tv6pm: 0. After check, didn't match packet.\n");
#endif */

	return NF_ACCEPT;
}

/***
 * This struct is used to register our interest for netfilter work
 ***/
static struct nf_hook_ops v4tv6_ops = 
{ { NULL, NULL }, v4tv6_hook, PF_INET, NF_IP_LOCAL_OUT, 0 };

/***
 * This is the function that kickstarts the module.
 ***/
static int __init v4tv6_init(void) 
{
	int err = 0;


	/* register with NF_IP_LOCAL_OUT */
	err = nf_register_hook(&v4tv6_ops);

	printk(KERN_INFO "v4tv6: IPv4 to IPv6 (outgoing) filter loaded (%d).\n"
			, err); 
	return err;
}

/***
 * This function stops the module.
 ***/
static void __exit v4tv6_exit(void) 
{ 
	/* unregister with NF_IP_LOCAL_OUT */
	nf_unregister_hook(&v4tv6_ops);
	printk(KERN_INFO "v4tv6: IPv4 to IPv6 (outgoing) filter unloaded.\n");
}


module_init(v4tv6_init)
module_exit(v4tv6_exit)
