/****************************************************************************
 *                                                                          *
 * Filename:    mdnspd-support.c                                            *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                      *
 * Created:     01/01/04                                                    *
 * Modified:    21/01/04                                                    *
 * Description: Contains the implmentation of the support functions from    *
 *              mdnspd-support.h                                            *
 *                                                                          *
 * (c) 2004 Tabitha Tipper, licensed under the GNU General Public License      *
 *     See the LICENSE file for details.                                    *
 ****************************************************************************/

/* FIXME: rewrite the scanning over encoded domain names to count better */
/* FIXME: add sempahore locking around the use of the socket */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <ctype.h>

/* For faffing the dnsrh device */
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include "../dnsrh/dnsrh.h"

#include "mdnspd-support.h"

/***
 * The thread function, does all the real forwarding, sending back, mangling
 * etc.
 *
 * void because of pthread-isms.
 ***/
void connection(void * data) { 
  int upChain = 0;		    /* socket for the upstream NS        */
  struct sockaddr_in upAddr;	    /* sockaddr for the upstream NS      */
  int * sendSock = NULL;	    /* socket for sending back down      */

  int addrlen = 0;		    /* length of a sockaddr_in struct    */

  unsigned char response[512];	    /* what the up server tells us       */
  int sendLen = 0;		    /* length of the sent request up     */
  int len = 0;			    /* length of response back from up   */
  int err = 0;			    /* generic error var                 */

  connectData * cd = (connectData *) data;
  sendSock = cd->serverSocket;
  addrlen = sizeof(struct sockaddr_in);

  /* Setup a socket and connect to the real NS */
  upChain = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (-1 == upChain)
    fprintf(stderr, "mdnspd: Error creating socket to upstream NS.\n");

  upAddr.sin_family = AF_INET;
  upAddr.sin_port = htons(53);
  inet_pton(AF_INET, cd->upstreamNS, &upAddr.sin_addr);

  /* Check for a 240.x.x.x reqest, if it exists look up its mapping and
   * rewrite the domain part.
   */
  if (isFake(cd->request, cd->reqLen)) { 
    err = changeToRealAddr(cd->request, &cd->reqLen);
  }

  /* Forward the request up to the upstream NS */
  sendLen = sendto(upChain, cd->request, cd->reqLen, 0, 
      (struct sockaddr *) &upAddr, addrlen);

  if (sendLen != cd->reqLen) { 
    fprintf(stderr, "MDNSPD: request forwarded to %s:%d possibly currupted.  Got %d bytes, sent %d\n", inet_ntoa(upAddr.sin_addr), 53, cd->reqLen, sendLen);
  }

  memset(response, '\0', 512); /* clear the request buffer */


  /* accept back from server, check for AAAA record and pass onto client */
  len = recvfrom(upChain, response, 512, 0, 
      (struct sockaddr *) NULL, &addrlen);

  /* If there weren't any answers from the query _AND_ its a type A
   * then upgrade to a AAAA and try again */
  if (0 == answers(response, len) 
      && 1 == qtypeA(cd->request, cd->reqLen)) {

    int sLen;
    int rLen;
    unsigned char newResponse[512];

    memset(newResponse, 0, 512);

    err = changeToTypeAAAA(cd->request, cd->reqLen);
    if (err != 0) { 

      /* Send/Recv with upstream */
      sLen = sendto(upChain, cd->request, cd->reqLen, 0,
	  (struct sockaddr *) &upAddr, addrlen);

      rLen = recvfrom(upChain, newResponse, 512, 0, 
	  (struct sockaddr *) NULL, &addrlen);

      /* In the event the AAAA query gets answers that the A didn't rewrite it
       * to be an A response and return it */
      if (1 == answers(newResponse, rLen)) {
	if (1 == recordAAAA(newResponse, rLen)) { /* check its AAAA */
	  err = changeToTypeA(newResponse, &rLen);
	}

	memset(response, 0, 512);
	memcpy(response, newResponse, rLen);
	len = rLen;
      }
    }
  }


  /* send back to the client */
  sem_wait(cd->sem);

  sendLen = sendto(*sendSock, response, len, 0, 
      (struct sockaddr *) &cd->clientAddr, addrlen);

  sem_post(cd->sem); 

  if (sendLen != len) { 
    fprintf(stderr, "mdnspd: request sent back down to %s:%d possibly currupted.  Got %d bytes, sent %d\n", inet_ntoa(cd->clientAddr.sin_addr), cd->clientAddr.sin_port, len, sendLen);
  }

  /* close struct */
  close(upChain);

  /* FREE MEMORY HERE of the _INCOMING_ struct */
  free(cd->request);
  free(cd->upstreamNS);
  free(data);
}

/***
 * Accepts a char array of a DNS response and an int of its length.
 *
 * Returns 1 if that response contains answering resource records, 0 if it
 * doesn't and -1 if the length is invalid. *
 ***/
int answers(unsigned char * in, int len) { 
  unsigned char lin[512];
  int answer = 0;
  int pos = 0;

  if (len >= 512 || len <= 12) { 
    return -1;
  }

  memset(lin, 0, 512);
  memcpy(lin, in, len);

  pos = 6; /* skip main header + questions field */

  answer += lin[pos++] & 0xFF00; /* get num. of answers */
  answer += lin[pos++] & 0x00FF;

  if (answer != 0) { /* there were answers */
    return 1;
  }

  return 0; /* No answers */
}

/*** 
 * Accepts a char pointer to a DNS query and an int of the length of that
 * query.
 *
 * Returns -1 if the len is invalid, 1 if any questions are a type A query, 
 * 0 if none of the questions are type A queries.
 ***/
int qtypeA(unsigned char * in, int len) { 
  int pos = 0;
  unsigned char lin[512];
  int type = 0;
  int questions = 0;
  int i = 0;

  if (len > 512 || len < 12) { /* check len is good */
    return -1;
  }

  memset(lin, '0', 512);
  memcpy(lin, in, len);

  pos += 4;

  questions += lin[pos++] & 0xFF00;
  questions += lin[pos++] & 0x00FF;

  pos += 6; /* end of headers now */

  for (i = 0; i < questions; i++) { 
    while (lin[pos++] != 0 && pos <= len);  /* Skip domain name */

    type += lin[pos++] & 0xFF00;
    type += lin[pos++] & 0x00FF;

    if (1 == type) { 
      return 1; /* Its true!  There are type A questions here! */
    }

    pos += 2; /* skip class */
  }

  return 0;
}

/***
 * Accepts a char pointer to a DNS query and an int for its length.  Modifies
 * it in place from a type A query (1) to an AAAA query (28).
 *
 * Changes ONLY THE QUESTIONS
 *
 * Returns 1 if the data was changed, 0 if not and -1 if the length is wrong.
 ***/
int changeToTypeAAAA(unsigned char * in, int len) { 
  int pos = 0;
  unsigned char lin[512];
  int type = 0;
  int questions = 0;
  int i = 0;
  int changed = 0;	/* set to 1 if the function has changed the data */

  if (len > 512 || len < 12) { /* check len is good */
    return -1;
  }

  memset(lin, '0', 512);
  memcpy(lin, in, len);

  pos += 4;

  questions += lin[pos++] & 0xFF00;
  questions += lin[pos++] & 0x00FF;

  pos += 6; /* end of headers now */

  for (i = 0; i < questions; i++) { 
    while (lin[pos++] != 0 && pos <= len);  /* Skip domain name */

    type += lin[pos++] & 0xFF00;
    type += lin[pos++] & 0x00FF;

    if (1 == type) { 
      changed = 1;

      lin[pos - 2] = 0;
      lin[pos - 1] = 28;
    }

    pos += 2; /* skip class */
  }

  memcpy(in, lin, len);

  return changed;
}

/***
 * Accepts a char array which is a DNS response, checks to see if it contains
 * AAAA response records.  Takes an int for the length of the character array
 *
 * Returns 1 if it is, 0 if it isn't, -1 if the length is wrong
 ***/
int recordAAAA(unsigned char * in, int len) { 
  int pos = 0;			/* current position in the request          */
  unsigned char lin[512];	/* local version of the request             */
  int questions = 0;		/* number of questions we need to jump      */
  int i = 0;

  if (len > 512 || len < 12) { /* Illegal length */
    return -1;
  }

  pos = 0;
  memset(lin, '0', 512);
  memcpy(lin, in, len);

  pos += 4; /* Jump past main headers */

  questions += lin[pos++] & 0xFF00; /* get number of questions */
  questions += lin[pos] & 0x00FF;

  pos += 6; /* Skip counters for RRs */

  for (i = 0; i < questions; i++) { 
    while (lin[pos++] != 0 && pos <= len);  /* Skip domain name */
    pos += 2; /* type/class */
  }

  while (pos < len) { /* Each reponse is */
    int rdLength = 0;
    int type = 0;

    while (lin[pos++] != 0 && pos <= len);  /* Skip domain name */

    /* 16 bits type, thats what we're looking for, a type 28 AAAA record */
    type += lin[pos++] & 0xFF00;
    type += lin[pos++] & 0x00FF;

    if (type == 28)
      return 1; /* EARLY RETURN HERE */

    /* 16 bits class */
    pos += 2;

    /* 32 bits TTL */
    pos += 4;

    /* 16 bits RDATA length, i.e. how much more to skip */
    rdLength += lin[pos++] & 0xFF00;
    rdLength += lin[pos++] & 0x00FF;

    /* variable RDATA */
    pos += rdLength;
  }

  return 0;
}

/*** 
 * Accepts a char array and modifies it in place, changing it from a AAAA
 * response to a A response, accepts an int for the length of the char array
 * too
 *
 * Changes all type AAAA answers, and all type AAAA questions
 *
 * Returns 1 if it modifies the input, 0 if not, -1 if length is wrong
 ***/
int changeToTypeA(unsigned char * in, int * len) {
  int pos = 0;		    /* Current position in the recieved buffer       */
  unsigned char lin[512];   /* the local copy of the input                   */
  int fLen = 0;		    /* the final length of the input                 */
  int questions = 0;	    /* number of questions being asked               */
  int rr = 0;		    /* number of resource records                    */
  int modified = 0;	    /* if record modified, 1 = yes, 0 = no           */
  int i = 0;
  int err = 0;

  if (*len > 512 || *len < 12) { /* check length */
    return -1;
  }

  pos = 0;
  fLen = *len;
  memset(lin, '0', 512);
  memcpy(lin, in, fLen);

  pos += 4; /* Jump past main headers */

  questions += lin[pos++] & 0xFF00; /* get number of questions */
  questions += lin[pos++] & 0x00FF;

  for (i = 0; i < 3; i++) { 
    err = 0;
    err += lin[pos++] & 0xFF00;
    err += lin[pos++] & 0x00FF;
    rr += err;
  }

  for (i = 0; i < questions; i++) { /* rewrites questions */
    int type = 0;

    while (pos <= fLen && lin[pos++] != 0);  /* Skip domain name */

    type += lin[pos++] & 0xFF00;
    type += lin[pos++] & 0x00FF;

    if (type == 28) { 
      lin[pos - 2] = (unsigned char) 0;
      lin[pos - 1] = (unsigned char) 1;
    }

    pos += 2; /* class */
  }

  for (i = 0; i < rr; i++) { /* Each reponse gets rewritten */ 
    char found = 0;
    int start = 0;
    int end = 0;

    unsigned char * domainName = NULL;
    /* char sectionLength = 0; */
    int dnLength = 0;
    int rdLength = 0;
    int type = 0;

    start = pos;

    /* Skip domain name */
    dnLength = getEncodedName(lin, fLen, NULL, pos);
    pos += dnLength;

    /* 16 bits type, thats what we're looking for, a type 28 AAAA record */
    type += lin[pos++] & 0xFF00;
    type += lin[pos++] & 0x00FF;

    /* replace record with an A record, get a new IP and write that in */
    if (28 == type) { 
      int i = 0;

      /* Actually bother copying the DN now */
      domainName = (unsigned char *) malloc(dnLength);
      for (i = 0; i < dnLength; i++) { 
	domainName[i] = lin[i + start];
      }

      found = 1;
    } 

    /* 16 bits class */
    pos += 2;

    /* 32 bits TTL */
    pos += 4;

    /* 16 bits RDATA length, i.e. how much more to skip */
    rdLength += lin[pos++] & 0xFF00;
    rdLength += lin[pos++] & 0x00FF;

    /* make sure we know where the end of this record *SHOULD* be so we 
     * can drag it back */
    end = (pos + rdLength);

    /* variable RDATA */
    if (1 == found) { 
      int i = 0;
      int handle = 0;
      int err = 0;
      ip_pair *ipp = (ip_pair *) malloc(sizeof(ip_pair));

      modified = 1;

      for (i = 0; i < rdLength; i++) { /* copy the V6 addr into ipp */
	ipp->v6[i] = lin[i + pos];
      }

      handle = open(DEVICE_PATH, O_RDWR); /* Add to hash, get V4 */
      err = ioctl(handle, IOCTL_ADD_V6, ipp);

      /* Copy new record back in */
      pos = start;

      for (pos = start; pos < fLen; pos++) { /* domain name */
	lin[pos] = in[pos];
	if (0 == lin[pos]) { 
	  break;
	}
      }

      pos++;  /* set the type 1 (A) */
      lin[pos++] = 1;

      for (i = 0; i < 6; i++) { /* class and TTL should be identical */
	lin[pos] = in[pos];
	pos++;
      }

      lin[pos++] = 0;	/* the new length, 4 bytes */
      lin[pos++] = 4;

      lin[pos++] = ipp->v4[0]; /* copy the new V4 address in */
      lin[pos++] = ipp->v4[1];
      lin[pos++] = ipp->v4[2];
      lin[pos++] = ipp->v4[3];

      /* shuffle the rest of the record back */
      for (i = end; i < *len; i++) { 
	lin[pos] = in[i];
	pos++;
      }

      fLen -= 12; /* remember our length is now 12 bytes shorter */
      free(domainName);
      free(ipp);
      close(handle);
    } 
    else { 
      pos += rdLength;
    }
  }

  memset(in, '0', 512); /* copy back from lin to in */
  memcpy(in, lin, fLen);

  *len = fLen;
  return modified;
}


/***
 * Takes a DNS query/response (req), its length (len) and a position (pos)
 * in that.  Starts reading through until it reaches the end of the domain
 * name.  Since I don't care it doesn't follow 192 pointers.
 *
 * If "out" is !NULL then it puts the final domain name (or up to the first
 * pointer) in out.
 *
 * It returns the number of bits long the domain name is, including pointer
 * address if one found.
 ***/
int getEncodedName(unsigned char * in, int len, unsigned char * out, int pos) { 
  unsigned char lin[512];
  char ch = 0;
  int start = 0;

  memset(lin, 0, 512);
  memcpy(lin, in, len);

  start = pos;

  ch = lin[pos++];

  /* According to RFC 1035, a pointer is a 2 octet field, (16 bits), where the
   * first two bits are 11, the rest (last 12) forming a pointer... */
  while (pos <= len) { 
    if (ch == 0) 
      break;

    if (ch == (char) 192)
      break;

    pos++;
    ch = lin[pos];
  }

  if (ch == (char) 192) { 
    pos++; /* skip pointer */
  }

  if (out != NULL) { /* write output if !NULL */
    out = (unsigned char *) malloc(sizeof(pos - start));
    memcpy(out, in, (pos - start));
  }

  return (pos - start);
}


/***
 * Takes in a char array of a DNS request and an int for its length.
 *
 * Returns 1 if its a "fake" i.e. is part of the 240.x.x.x range.  Returns 0
 * if its not.  returns -1 if the length is invalid.
 ***/
int isFake(unsigned char * req, int len) { 
  int pos = 0;	      /* position in req */
  int numq = 0;	      /* number of questions */
  int i = 0;

  if (len > 512 || len < 12) { 
    return -1;
  }

  pos = 4;

  numq += req[pos++] & 0xFF00; /* number of questions */
  numq += req[pos++] & 0x00FF;

  pos += 6; /* skip answers + RRs, leave on 1st block of 1st question */

  for (i = 0; i < numq; i++) { /* for each question */
    int sec = 0; /* size of each section for encoded domain names */

    /* first char (skip length) is a digit, could be in-addr.arpa */
    if (isdigit(req[(pos + 1)])) {
      int j = 0;
      int lpos = 0;
      unsigned char num[4];
      memset(num, '\0', 4);

      for (j = 0; j < 3; j++) { /* read first three */
	sec = req[pos++];
	pos += sec;
      }

      sec = req[pos++];
      for (j = 0; j < sec; j++) { /* get the last chunk */
	num[lpos] = req[pos];
	lpos++;
	pos++;
      }

      if (req[pos] != 0) { /* check thats not the end of the name */
	pos++;

	if (atoi(num) == 240) { /* number right, check for i-a.a */
	  if (0 != strncmp("in-addr", &req[pos], 7)) {
	    if (DEBUG) { 
	      fprintf(stderr, 
		  "mdnspd: error matching in-addr got '%7.7s' instead.\n", 
		  &req[pos]);
	    }
	    break;
	  }

	  pos += 8; /* skip in-addr + length */

	  if (0 != strncmp("arpa", &req[pos], 4)) { 
	    break;
	  }

	  /* well its valid, return */
	  return 1;
	} 
      }
    }

    else { /* skip encoded domain name */
      int gen = getEncodedName(req, len, NULL, pos);
      pos += gen;
    }

    pos += 4; /* skip type/class */
  }

  return 0;
}


/***
 * int-to-hex
 *
 * Accepts a char and converts it from its charness to the value of that
 * symbol in hex notation as a decimal.
 * 
 * Return -1 on error, 0-15 in other cases.
 ***/
unsigned char ith(unsigned char in) {       
  switch (in) {
    case 0:
      return '0';
    case 1:
      return '1';
    case 2:
      return '2';
    case 3:
      return '3';
    case 4:
      return '4';
    case 5:
      return '5';
    case 6:
      return '6';
    case 7:
      return '7';
    case 8:
      return '8';
    case 9:
      return '9';
    case 10:
      return 'a';
    case 11:
      return 'b';
    case 12:
      return 'c';
    case 13:
      return 'd';
    case 14:
      return 'e';
    case 15:
      return 'f';
    default:
      return -1;
  }

  return -1;
}

/***
 * Accepts a char array of a DNS request and an int of its length
 * Rewrites the address from a 240.0.0.0/24 to its real v6 address.
 *
 * Also changes the question types to AAAA from A.
 *
 * Returns 1 on success, 0 on failure, -1 if the length is invalid
 ***/
int changeToRealAddr(unsigned char * req, int * len) { 
  ip_pair * ipp = NULL;	      /* data structure with addresses in */
  unsigned char lin[512];     /* local copy of data */
  int nLen = 0;		      /* New Length (for req) */
  int pos = 0;		      /* current postion in req */
  int numq = 0;		      /* number of questions to modify */
  int changed = 0;	      /* 0 if no changes 1 if input modified */
  int i = 0;
  int j = 0;

  if (*len > 512 || *len < 12) {
    return -1;
  }

  memset(lin, 0, 512);
  memcpy(lin, req, *len);

  nLen = *len;

  ipp = (ip_pair *) malloc(sizeof(ip_pair)); 

  /* No need to change questions, should already be PTR
     changeToTypeAAAA(lin, nLen); */

  pos = 4;

  numq += lin[pos++] & 0xFF00; /* number of questions */
  numq += lin[pos++] & 0x00FF;

  pos += 6; /* skip answers and RR sections */

  for (i = 0; i < numq; i++) { 
    int start = pos;
    int handle = 0;
    int err = 0;

    for (j = 3; j >= 0; j--) { /* get the quads */
      int k = 0;
      unsigned char num[4];
      int sec = 0;

      sec = lin[pos++];

      for (k = 0; k < sec; k++) {
	num[k] = lin[pos];
	pos++;
      }

      ipp->v4[j] = atoi(num);
    }

    if (DEBUG) { 
      fprintf(stderr, "mdnspd: quads: %d.%d.%d.%d\n", 
	  ipp->v4[0], 
	  ipp->v4[1], 
	  ipp->v4[2], 
	  ipp->v4[3]);
    }

    pos += 15; /* skip "7in-addr4arpa0" */
    pos += 4; /* skip type/class */

    handle = open(DEVICE_PATH, O_RDWR); /* Add to hash, get V4 */
    err = ioctl(handle, IOCTL_CHECK_FOR_V6, ipp);


    /* need to use nibble format: e.g. 
     * dig 3.5.2.0.0.0.0.0.0.0.0.0.0.0.0.0.3.0.0.0.2.0.0.0.1.0.0.0.0.c.e.f.ip6.arpa PTR
     */
    if (0 == err) { /* there was a v6 match */
      int cpos = start;
      int end = 0;

      changed = 1;

      if (DEBUG) { 
	fprintf(stderr, 
	    "mdnspd: err was 0, there was a match. start is %d\n", start);

	for (i = 16; i >= 0; i--) { /* DEBUGGING */
	  fprintf(stderr, "%d: all %u, top: %u, bottom: %u\n", 
	      i, 
	      ipp->v6[i], 
	      (ipp->v6[i] & 0xF0) >> 4, 
	      (ipp->v6[i] & 0x0F)
	      );
	}

	fprintf(stderr, "real thing: \n");
      }

      for (i = 15; i >= 0; i--) { /* 16 chars worth of data, 32 chunks */
	lin[cpos++] = 1;
	lin[cpos++] = ith((ipp->v6[i] & 0x0F));
	lin[cpos++] = 1;
	lin[cpos++] = ith(((ipp->v6[i] & 0xF0) >> 4));
      } 

      lin[cpos++] = 3;
      lin[cpos++] = 'i';
      lin[cpos++] = 'p';
      lin[cpos++] = '6';
      lin[cpos++] = 4;
      lin[cpos++] = 'a';
      lin[cpos++] = 'r';
      lin[cpos++] = 'p';
      lin[cpos++] = 'a';
      lin[cpos++] = 0;

      cpos++;
      /* lin[cpos++] = 38; * Type is A6, RFC 2874 has details */
      lin[cpos++] = 12; /* actually its PTR, seems RFC 2874 isn't working yet */
      cpos++;
      lin[cpos++] = 1;

      end = cpos;

      /* copy stuff back from the original */
      for (i = pos; i < *len; i++) { 
	lin[cpos] = req[i];
	cpos++;
      }

      /* get ready for the next question, adjust pos to start of next Q */
      pos = end;

      /* adjust nLen */
      nLen += (cpos - start);

      if (DEBUG) { 
	for (i = 0; i < (cpos + 4); i++) { 
	  fprintf(stderr, "%d:  '%c' [ %d ]\n", i, lin[i], lin[i]);
	}
      }

    } 
    else { 
      if (DEBUG) { 
	fprintf(stderr, "mdnspd: err was %d\n", err);
      }
    }
  }

  /* copy lin to req and nLen to len if something changed */
  if (changed) { 
    memset(req, 0, 512);
    memcpy(req, lin, nLen);
    *len = nLen;
    return 1;
  }

  return 0;
}
