/****************************************************************************
 *                                                                          *
 * Filename:    mdnspd-support.h                                            *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                      *
 * Created:     01/01/04                                                    *
 * Modified:    01/01/04                                                    *
 * Description: Defines the functions to identify and change AAAA records   *
 *              to type A records (IPv6 resposes to IPv4) consistently      *
 *                                                                          *
 * (c) 2004 Tabitha Tipper, licensed under the GNU General Public License      *
 *     See the LICENSE file for details.                                    *
 ****************************************************************************/

#ifndef mdnspd_support_h
#define mdnspd_support_h

/* Networking headers */
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

/* sempahore */
#include <semaphore.h> 

#define PORTNUM 53
#define LHOST localhost
#define DEVICE /dev/dnsrh
#define DEFAULT_UPSTREAM "10.0.0.253"
#define DEVICE_PATH "/dev/dnsrh"

/* set to 1 if you want debugging output, 0 otherwise */
#define DEBUG 1

typedef struct _connectData { 
  sem_t * sem;   
  int * serverSocket;		    /* The socket of the server */
  unsigned char * request;	    /* request from client */
  int reqLen;			    /* length of the request of the client */
  char * upstreamNS;		    /* address of the upstream (real) NS */
  struct sockaddr_in clientAddr;    /* info about client */
} connectData;

/***
 * The thread function, does all the real forwarding, sending back, mangling
 * etc.
 *
 * void because of pthread-isms *
 ***/
void connection(void * data);

/***
 * Accepts a char array of a DNS response and an int of its length.
 *
 * Returns 1 if that response contains answering resource records, 0 if it
 * doesn't and -1 if the length is invalid. *
 ***/
int answers(unsigned char * in, int len);

/*** 
 * Accepts a char pointer to a DNS query and an int of the length of that
 * query.
 *
 * Returns -1 if the len is invalid, 1 if its a type A query, 0 if its not
 ***/
int qtypeA(unsigned char * in, int len);

/***
 * Accepts a char pointer to a DNS query and an int for its length.  Modifies
 * it in place from a type A query (1) to an AAAA query (28).
 *
 * Returns 1 on success, 0 on failure and -1 if the length is wrong.
 ***/
int changeToTypeAAAA(unsigned char * in, int len);

/***
 * Accepts a char array which is a DNS response, checks to see if it contains
 * AAAA response records.  Takes an int for the length of the character array
 *
 * Returns 1 if it is, 0 if it isn't, -1 if the length is wrong
 ***/
int recordAAAA(unsigned char * in, int len);

/* Accepts a char array and modifies it in place, changing it from a AAAA
 * response to a A response, accepts an int for the length of the char array
 * too
 *
 * Returns 1 on success, 0 on failure, -1 if length is wrong
 */
int changeToTypeA(unsigned char * in, int * len);

/***
 * Takes a DNS query/response (req), its length (len) and a position (pos)
 * in that.  Starts reading through until it reaches the end of the domain
 * name.  Since I don't care it doesn't follow 192 pointers.
 *
 * If "out" is !NULL then it puts the final domain name (or up to the first
 * pointer) in out.
 *
 * It returns the number of bits long the domain name is, including pointer
 * address if one found.
 ***/
int getEncodedName(unsigned char * req, int len, unsigned char * out, int pos);

/***
 * Takes in a char array of a DNS request and an int for its length.
 *
 * Returns 1 if its a "fake" i.e. is part of the 240.x.x.x range.  Returns 0
 * if its not.  returns -1 if the length is invalid.
 ***/
int isFake(unsigned char * req, int len);

/***
 * int-to-hex
 *
 * Accepts a char and converts it from its charness to the value of that
 * symbol in hex notation as a decimal.
 * 
 * Return -1 on error, 0-15 in other cases.
 ***/
unsigned char ith(unsigned char in);

/***
 * Accepts a char array of a DNS request and an int of its length
 * Rewrites the address from a 240.0.0.0/24 to a read address
 *
 * Returns 1 on success, 0 on failure, -1 if the length is invalid
 ***/
int changeToRealAddr(unsigned char * req, int * len);

#endif
