/****************************************************************************
 *                                                                          *
 * Filename:    test-client.c                                               *
 * Author:      Tabitha Tipper <t.tipper@lancaster.ac.uk>                      *
 * Created:     17/12/03                                                    *
 * Modified:    17/12/03                                                    *
 * Description: Sends a DNS request up to a specified server:port and       *
 *              prints the response (if any).                               *
 *                                                                          *
 *              test-client <host> <server> <port>                          *
 *                                                                          *
 ****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

/* returns the given integer as a char */
char itoc(int in) { 
  char binny = '0';
  binny = ((binny | 0xFF) & in);
  return binny;
}

/* prints out a single char as binary bits */
void printSingleBinary(unsigned char input) { 
   printf("%d%d%d%d%d%d%d%d", ((input & 0x80) >> 7), ((input & 0x40) >> 6), 
      ((input & 0x20) >> 5), ((input & 0x10) >> 4), ((input & 0x08) >> 3), 
      ((input & 0x04) >> 2), ((input & 0x02) >> 1), (input & 0x01)
   );
}

/* returns 0 if input is a dotted quad, 1 otherwise */
int dottedQuad(char * question) {
  char * sect;
  int i = 0;
  int result = 1;
  char local[67];

  strncpy(local, question, 67);
  sect = (char *) strtok(local, ".");

  while (sect != NULL) {
    for (i = 0; i < strlen(sect); i++) {
      if (isdigit(sect[i]) == 0) {
	result = 0;
	break;
      }
    }
    if (result == 0) { break; }
    sect = (char *) strtok(NULL, ".");
  }

  return result;
}

int main(int argc, char * argv[]) { 

  int err = 0;
  int i = 0;

  char host[100];	      /* holds the args */
  char server[17];
  char portString[5];
  int port;

  int sock = 0;	      /* the network sock */
  struct sockaddr_in addy;    /* details of the NS */

  char request[512];	      /* the request itself */
  int currPos = 0;	      /* current postion in request */
  char encdom[512];	      /* the encoded domain name */
  char * section;	      /* a temp section of strings */


  /* Deal with command line args */
  if (argc < 2) { 
    fprintf(stderr, "Incorrect commandline args: test-client <host> <server> [<port>] \n");
    return 1;
  }

  memset(host, '\0', 100);
  memset(server, '\0', 17);
  memset(portString, '\0', 5);

  strncpy(host, argv[1], 100);
  strncpy(server, argv[2], 17);

  if (argc >= 3) { /* optional port number */
    strncpy(portString, argv[3], 5);
  } else { 
    strncpy(portString, "53", 5);
  }

  port = atoi(portString);

  /* Setup sock */
  sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if (-1 == sock) {
    fprintf(stderr, "test-client: Error creating sock.\n");
    return 1;
  }

  addy.sin_family = AF_INET;
  addy.sin_port = htons((unsigned short) port); /* need "server" lookup */
  err = inet_aton(server, &addy.sin_addr);
  if (0 == err) { 
    fprintf(stderr, "test-client: Error converting host (%s) to sin_addr: %s", host, strerror(errno));
    return 1;
  }

  i = sizeof(struct sockaddr);
  err = connect(sock, (struct sockaddr *) &addy, i);
  fprintf(stderr, "test-client: connected to %s:%d.\n", server, port);

  /* Construct request */
  memset(request, 0, 512);

  request[0] = (1 & 0xFF00);    /* set identification */
  request[1] = (1 & 0x00FF);

  request[6] = (1 & 0x00FF);	/* set number of questions */

  currPos = 12;	/* We're after the headers now */

  /* Encode domain name */
  if (dottedQuad(host)) { /* check and fix dotted quad name */
    int a, b, c, d;
    memset(section, '\0', 100);

    section = strtok(host, ".");
    d = atoi(section);
    section = strtok(NULL, ".");
    c = atoi(section);
    section = strtok(NULL, ".");
    b = atoi(section);
    section = strtok(NULL, ".");
    a = atoi(section);
    section = strtok(NULL, ".");

    sprintf(host, "%d.%d.%d.%d.in-addr.arpa", 
	a, b, c, d);
  }
  printf("host: %s, server: %s:%d\n", host, server, port);

  memset(encdom, 0, 512);

  printf("Encoding request.\n");
  section = strtok(host, "."); /* tokenise + First Section */
  while (section != NULL) { /* append each section */
    printf("(%d)%d: %s\n", 
	strlen(encdom), (int) itoc(strlen(section)), section);
    if (0 == strlen(encdom)) { 
      sprintf(encdom, "%c%s", itoc(strlen(section)), section);
    } else { 
      printf("encdom not empty (%s).\n", encdom);
      sprintf(encdom, "%s%c%s", encdom, itoc(strlen(section)), section);
    }
    section = strtok(NULL, ".");
  }
  sprintf(encdom, "%s%c", encdom, itoc(0)); /* put a 0 at the end */

  printf("length of encdom: %d, contents: '%s'.\n", strlen(encdom), encdom);

  /* append from after header to request, remember to include the 0 */
  for (i = 0; i < (strlen(encdom) + 1); i++) {
    request[(i + 12)] = encdom[i];
    printf("%d/%d: '%c'/'%d'\n", i, i + 12, encdom[i], encdom[i]);
    currPos++;
  }

  /* append Type and Class values */
  currPos++;
  request[currPos++] = 2 & 0x0000FFFF; /* NS request */

  currPos++;
/*  request[currPos] = request[currPos] | 0x01;   * Class (INternet) */
  request[currPos] = 1 & 0x0000FFFF;
  currPos++;
  
  /* print out the request (header, domain, type/class) */
  printf("Header (1-12 of %d): \n", currPos);
  err = 0;
  for (i = 0; i < 12; i++) { /* header */
    /* printf("%d2.2: ", i); */
    printSingleBinary(request[i]);
    printf(" ");
    if (1 == err) { 
      printf("\n");
      err = 0;
    } else { 
      err = 1;
    }    
  }
  printf("\nDomain (all bar last 4 bytes of %d): \n", currPos);

  for (i = 12; i < (currPos - 4); i++) { /* domain */
    printf("%d: char: '%c'\tint: '%d', \n", i, request[i], (int) request[i]);
  }

  printf("\nType/Class (last 4 bytes of %d): \n", currPos);
  err = 0;
  for (i = (currPos - 4); i < currPos; i++) {  /* type/class */
    /* printf(" %d: ", i); */
    printSingleBinary(request[i]);
    printf(" ");
    if (1 == err) { 
      printf("\n");
      err = 0;
    } else { 
      err = 1;
    } 
  }

  /* send the request to the server */
  err = send(sock, request, currPos, 0);

  if (-1 == err) { 
    fprintf(stderr, "test-client: Error sending data through sock.\n");
    return 1;
  }
 
  return 0;
}
